﻿
local mod = klhtm
local me = { }
mod.my = me

--[[ 
KTM_My.lua

This file stores data that specifically relates to the current player, and methods that change these data sets.
Most of the data sets are updated by polling, that is frequently recalculating their values.

me.states		- 	a set of flags, e.g. "player in combat" / "afk" / "feigning death"
me.spellranks	-	list of ranks the player has for the important threat causing spells / abilities
me.mods			-	currently active modifiers to threat / abilities from talents, gear, buffs, etc
me.globalthreat	-	buffs that affect all threat the player generates
]]

----------------------------------------------------------------------------------------

me.lastupdatetime = 0.0	-- return value of GetTime()
me.longupdateinterval = 1.0 -- at least this time in seconds will pass between long updates
me.feigndeathresisttime = 0 -- return value of GetTime()
me.lastfadevalue = 0

-----------------------------------------
--    Special Methods from Core.lua    --
-----------------------------------------

-- The "UI_ERROR_MESSAGE" event is used to detect Feign Death resists.
me.myevents = { "UI_ERROR_MESSAGE", "CHAT_MSG_SPELL_PERIODIC_SELF_DAMAGE", "CHAT_MSG_SPELL_AURA_GONE_SELF", 
	"CHAT_MSG_SPELL_ITEM_ENCHANTMENTS", "KLHTM_STATECHANGE_FADE", }

me.onevent = function()

	local ability

	if event == "UI_ERROR_MESSAGE" then
		
		if arg1 == ERR_FEIGN_DEATH_RESISTED then
			mod.out.printtrace("feigndeathdebug", "Feign Death Resist message intercepted.")
			me.feigndeathresisttime = GetTime()
		end
		
	elseif event == "CHAT_MSG_SPELL_PERIODIC_SELF_DAMAGE" then
		
		-- Check for burning adrenaline
		_, _, ability = string.find(arg1, mod.string.get("combatlog", "debuffstart"))
		
		if ability == mod.string.get("spell", "burningadrenaline") then
			me.setstate("burningadrenaline", true)
		end
				
	elseif event == "CHAT_MSG_SPELL_AURA_GONE_SELF" then
		
		-- Buff / Debuff fades
		_, _, ability = string.find(arg1, mod.string.get("combatlog", "buffend"))
		if ability then
		
			if ability == mod.string.get("spell", "burningadrenaline") then
				me.setstate("burningadrenaline", false)
				
			elseif ability == mod.string.get("spell", "arcaneshroud") then
				me.setstate("arcaneshroud", false) 
			
			elseif ability == mod.string.get("spell", "fade") then
				-- retract fade
				mod.table.raidthreatoffset = mod.table.raidthreatoffset + me.lastfadevalue
				me.setstate("fade", false)
			end 
		end
	
	elseif event == "CHAT_MSG_SPELL_ITEM_ENCHANTMENTS" then
	
		mod.out.printtrace("rockbiter", "Got this event text: " .. arg1)
	
		-- only bother for shamans, who have rockbiter
		if UnitClass("player") == mod.string.get("class", "shaman") then
			
			-- check for weapon enchant
			_, _, ability = string.find(arg1, mod.string.get("combatlog", "weaponbuff"))
			
			mod.out.printtrace("rockbiter", "enchantment is " .. tostring(ability))
			
			-- check for rockbiter
			if ability and string.find(ability, mod.string.get("spell", "rockbiter")) then
			
				mod.out.printtrace("rockbiter", "identified rockbiter")
				
				-- get rank
				local rank
				_, _, rank = string.find(ability, ".-(%d+).-")
				
				mod.out.printtrace("rockbiter", "rank = " .. rank)
				
				-- set rank in mods
				me.mods.shaman.rockbiter = tonumber(rank)
			
			else
			
				-- added an enchant that was not rockbiter
				me.mods.shaman.rockbiter = 0
			end
		end
	end

end

--[[  
Special onupdate() function, called from Core.lua
me.states		-	frequent update
me.spellranks	-	long time update
me.mods			-	long time update
me.globalthreat	-	frequent update
]]
me.onupdate = function()
	
	-- check for long updates
	local timenow = GetTime()
	
	if timenow > me.lastupdatetime + me.longupdateinterval then
	
		me.lastupdatetime = timenow
		me.redospellranks()
		me.redomods()
	end
	
	-- short updates
	me.redostates()
	me.redoglobalthreat()
	
	-- check status of rockbiter enchant for Shaman, i.e. whether it has run out.
	if GetWeaponEnchantInfo() == nil then
		me.mods.shaman.rockbiter = 0
	end
end

----------------------------------------------------------------------------------------

----------------------------
--	  General Methods     --
----------------------------


--[[ 
mod.my.ability(name, value)
Returns the current value for some property of your spells and abilities.
	<name> is the name of the ability, e.g. "Heroic Strike", "Feint". It is localised.
	<value> is the name of the property, e.g. "threat", "multiplier", "rage". These are the indexes of values in
the mod.data.abilities structure.
	This method will use me.spellranks to get the data that applies to your specific rank of the ability. Then
it will apply any modifiers from me.mods that are appropriate.
]]
me.ability = function(name, parameter)
		
	if mod.data.abilities[name] == nil then
		mod.out.printtrace("badabilityname", string.format("No ability |cffffff00%s|r found.", tostring(name)))
		return 0
	end
	
	local value
	
	-- First search for a spell rank override
	if mod.data.abilities[name][me.spellranks[name]] and mod.data.abilities[name][me.spellranks[name]][parameter] then
		value = mod.data.abilities[name][me.spellranks[name]][parameter]
	else
		value = mod.data.abilities[name].default[parameter]
	end
	
	-- now look for specific mods
	if name == mod.string.get("spell", "sunder") then
		if parameter == "rage" then
			value = value + me.mods.warrior.sundercost
		elseif parameter == "threat" then
			value = value * me.mods.warrior.sunderthreat
		end
		
	elseif name == mod.string.get("spell", "heroicstrike") then
		if parameter == "rage" then
			value = value + me.mods.warrior.heroicstrikecost
		end
		
	elseif name == mod.string.get("spell", "feint") then
		if parameter == "threat" then
			value = value + me.mods.rogue.feintthreat
		end
		
	elseif name == mod.string.get("spell", "maul") then
		if parameter == "rage" then
			value = value + me.mods.druid.ferocity
		end
		
	elseif name == mod.string.get("spell", "swipe") then
		if parameter == "rage" then
			value = value + me.mods.druid.ferocity
		end	
	end
	
	return value
end

--[[
mod.my.testthreat()
Print out your threat properties for debug purposes.
]]
me.testthreat = function()
		
	-- 1) Print out spell ranks
	local key
	local value
	
	for key, value in me.spellranks do
		mod.out.print(string.format(mod.string.get("print", "data", "abilityrank"), key, value))
	end
	
	-- 2) Print out global threat
	mod.out.print(string.format(mod.string.get("print", "data", "globalthreat"), me.globalthreat.value))
	for key, value in me.globalthreat.modifiers do
		if value.isactive == false then
			break
		end
		
		mod.out.print(string.format(mod.string.get("print", "data", "globalthreatmod"), value.reason, value.value))
	end

	-- 3) Print out threat for specific abilities
	if UnitClass("player") == mod.string.get("class", "priest") then
		mod.out.print(string.format(mod.string.get("print", "data", "multiplier"), mod.string.get("class", "priest"), mod.string.get("talent", "silentresolve"), 1.0 + me.mods.priest.silentresolve))
		mod.out.print(string.format(mod.string.get("print", "data", "multiplier"), mod.string.get("class", "priest"), SPELL_SCHOOL5_CAP, me.mods.priest.shadowaffinity))
	
	elseif UnitClass("player") == mod.string.get("class", "warlock") then
		mod.out.print(string.format(mod.string.get("print", "data", "setactive"), mod.string.get("sets", "nemesis"), 8, mod.out.booltostring(me.mods.warlock.nemesis)))
	
	elseif UnitClass("player") == mod.string.get("class", "mage") then
		mod.out.print(string.format(mod.string.get("print", "data", "setactive"), mod.string.get("sets", "netherwind"), 3, mod.out.booltostring(me.mods.mage.netherwind)))
		mod.out.print(string.format(mod.string.get("print", "data", "multiplier"), mod.string.get("class", "mage"), SPELL_SCHOOL6_CAP, 1.0 + me.mods.mage.arcanethreat))
		mod.out.print(string.format(mod.string.get("print", "data", "multiplier"), mod.string.get("class", "mage"), SPELL_SCHOOL4_CAP, 1.0 + me.mods.mage.frostthreat))
		mod.out.print(string.format(mod.string.get("print", "data", "multiplier"), mod.string.get("class", "mage"), SPELL_SCHOOL2_CAP, 1.0 + me.mods.mage.firethreat))
		
	elseif UnitClass("player") == mod.string.get("class", "paladin") then
		mod.out.print(string.format(mod.string.get("print", "data", "multiplier"), mod.string.get("class", "paladin"), mod.string.get("print", "data", "holyspell"), me.mods.paladin.righteousfury))
		mod.out.print(string.format(mod.string.get("print", "data", "healing"), mod.data.threatconstants.healing * me.mods.paladin.healing))
		
	elseif UnitClass("player") == mod.string.get("class", "warrior") then
		mod.out.print(string.format(mod.string.get("print", "data", "multiplier"), mod.string.get("class", "priest"), mod.string.get("spell", "sunder"), me.mods.warrior.sunderthreat))
		
	elseif UnitClass("player") == mod.string.get("class", "druid") then
		mod.out.print(string.format(mod.string.get("print", "data", "multiplier"), mod.string.get("class", "druid"), mod.string.get("talent", "tranquility"), me.mods.druid.tranquilitythreat))
		
	elseif UnitClass("player") == mod.string.get("class", "shaman") then
		mod.out.print(string.format(mod.string.get("print", "data", "healing"), mod.data.threatconstants.healing * me.mods.shaman.healing))
		
		if me.mods.shaman.rockbiter > 0 then
			mod.out.print(string.format(mod.string.get("print", "data", "rockbiter"), me.mods.shaman.rockbiter, UnitAttackSpeed("player") * mod.data.rockbiter[me.mods.shaman.rockbiter]))
		end
	end
end


----------------------------------------------------------------------------------------

----------------------------
--	  	  States          --
----------------------------

--[[      
each state is represented by a key-value pair. The key is a string, which is a description of the state,
e.g. "incombat", "afk". The value is a list with properties 
	["value"], boolean, whether the state is on or off
	["lastchange"], return value of GetTime() (seconds + decimal), when the value last changed
	["duration"], OPTIONAL, known duration of the state. This is used to turn the state off just in case
				  the normal mechanism to detect it does not work.
				
To set a state's value, call the method mod.my.setstate(<string name>, <boolean value>). If this causes the
state to change, the method will return non-nil, and an event KLHTM_STATECHANGE_<NAME> will be raised.
]]  
me.states = 
{
	["feigndeath"] = 
	{
		["value"] = false,
		["lastchange"] = 0,
	},
	["arcaneshroud"] = -- from Fetish of the Sandreaver
	{
		["value"] = false,
		["lastchange"] = 0,
		["duration"] = 15.0
	},
	["incombat"] = 
	{
		["value"] = false,
		["lastchange"] = 0,
	},
	["playercharmed"] = -- whether you have been mind controlled
	{
		["value"] = false,
		["lastchange"] = 0,
	},
	["burningadrenaline"] = 
	{
		["value"] = false,
		["lastchange"] = 0,
		["duration"] = 20.0,
	},
	["afk"] = 
	{
		["value"] = false,
		["lastchange"] = 0,
	},
	["fade"] = 
	{
		["value"] = false,
		["lastchange"] = 0,
		["duration"] = 10.0,
	},
}


--[[ 
mod.my.setstate(state, value)
Sets the value of one of the state variables above.
<state> is a string, a key to the me.states list, e.g. "afk", "incombat", etc.
<value> is a boolean, true or false.
Return: true if the set represents a change, otherwise nil
]]
me.setstate = function(state, value)
	
	-- verify <state>
	if me.states[state] == nil then
		mod.out.printtrace("statebadkey", string.format("There is no state |cffffff00%s|r.", tostring(state)))
		return
	end
	
	-- update
	if me.states[state].value ~= value then
		me.states[state].value = value
		me.states[state].lastchange = GetTime()
		
		return true
	end

end

--[[
me.redostates()
Checks for changes to state variables that we poll for (e.g. "incombat" / "ischarmed").
Also checks for states that have run out, but whose end events were not detected (e.g. burning adrenaline). This
last should not be relied upon though.
]]
me.redostates = function()
	
	-- hunter: check for FD
	if UnitClass("player") == mod.string.get("class", "hunter") then
		
		-- Check for feign death debuff
		local currentfd = mod.data.isbuffpresent("Interface\\Icons\\Ability_Rogue_FeignDeath")
		
		if me.setstate("feigndeath", currentfd) and (currentfd == true) then
			
			-- first check for resist.
			if math.abs(me.feigndeathresisttime - GetTime()) < 1.0 then
				--resisted. Nothing happens
				mod.out.printtrace("feigndeathdebug", "Feign Death was resisted!")
				
			else -- wipe threat
			
				mod.combat.event.threat = - mod.table.getraidthreat()
				mod.combat.event.rage = 0
				mod.combat.event.damage = 0
				mod.combat.event.name = "Feign Death"
				mod.combat.event.hits = 0
				
				mod.combat.addattacktodata(mod.string.get("threatsource", "threatwipe"), mod.combat.event)
				
				-- threat wipe: add to raid threat offset, not total threat.
				mod.table.raidthreatoffset = mod.table.raidthreatoffset + mod.combat.event.threat
				--mod.combat.addattacktodata(mod.string.get("threatsource", "total"), mod.combat.event)
				
				KLHTM_RequestRedraw("self")
			end
			
		end
	end
 
	-- update charmed state
	if UnitIsCharmed("player") == 1 then
		me.setstate("playercharmed", true)
	else
		me.setstate("playercharmed", false)
	end

	-- update combat state
	if UnitAffectingCombat("player") == 1 then
		
		if me.setstate("incombat", true) then
			-- player has just joined combat
			mod.out.printtrace("incombat", "You entered combat.")
			
			mod.table.resetraidthreat()
			
			local key
			local value
		
			for key, value in mod.combat.recentattacks do
				mod.table.raidthreatoffset = mod.table.raidthreatoffset + value[2]
			end	
		end
	
	else
		if (me.states.playercharmed.value == true) or (GetTime() - me.states.playercharmed.lastchange < 2.0) then
			me.setstate("incombat", true) -- should not be a change
		else
			if me.setstate("incombat", false) then
				mod.out.printtrace("incombat", "You left combat.")
				me.lastfadevalue = 0 -- stop fade
			end 
		end
	end
	
	-- update states with a "duration" parameter. They are all designed to deactivate on their own, but e.g.
	-- burning adrenaline isn't working, so we put in this mechanism as a backup
	local state
	local data
	
	for state, data in me.states do
		if data.duration and (data.value == true) and (GetTime() > data.lastchange + data.duration + 1.0) then
			me.setstate(state, false)
			mod.out.printtrace("statetimeout", "Deactivated the state " .. state .. " because it had passed its duration.")
		end
	end 
end

----------------------------------------------------------------------------------------

----------------------------
--      Spell Ranks       --
----------------------------

--[[
	A collection of key-value pairs. The key is a string, the name of a spell / ability, e.g. "Sunder Armor". It
is localised. The value is an integer, the rank of that spell you have.
	The only spells recorded are those which are cause or remove threat by known amounts, i.e. Sunder Armor, 
Revenge, Maul, Feint, etc.
	We want to know the what rank of the spell the player has, because most abilities have different numbers for 
different spell ranks. e.g. Heroic Strike does more damage and threat for each new rank. Some players will have the
extra rank from AQ20, some will not.
	These ranks will probably not change over the course of a session, but we still poll them for simplicity.
]]
me.spellranks = { } 

--[[
me.redospellranks = function()
Redetermines the ranks of all special threat abilities you have.
Heap Allocation: 1 gfind for each ability you have. The maximum possible is 5, for a warrior. Each string.gfind is
64 bytes, so that's a maximum 320 bytes.
]]
me.redospellranks = function()
	
	local index = 0
	local name
	local rankstring
	local rank
	
	while true do
		index = index + 1
		name, rankstring = GetSpellName(index, "spell")
		
		if name == nil then
			break
		end
		
		if mod.data.abilities[name] then
			rank = 0
			
			for rank in string.gfind(rankstring, mod.string.get("misc", "spellrank")) do
				me.spellranks[name] = rank
			end
		end
	end
end


----------------------------------------------------------------------------------------

---------------------------------------
--    Threat And Ability Modifers    --
---------------------------------------

me.mods = 
{
	["warrior"] = 
	{
		["defiance"] = 0.0, -- modifier to global threat. e.g. "+0.15" for 5 points.
		["sunderthreat"] = 1.0, -- multiplier. e.g. "1.15" for 8/8 Might.
		["sundercost"] = 0.0, -- modifier of rage cost. e.g. "-3" for 3/3 improved sunder talent.
		["heroicstrikecost"] = 0.0, -- modifier of rage cost. e.g. "-3" for 3/3 improved heroic strike talent.
		["impale"] = 0.0 -- critical strike bonus damage for abilities. e.g. "0.2" for 2/2 impale talent.
	},
	["rogue"] = 
	{
		["feintthreat"] = 1.0, -- multiplier. e.g. "1.25" for 5/8 Bloodfang.
	},
	["druid"] = 
	{
		["feralinstinct"] = 0.0, -- defiance for druids.
		["savagefury"] = 1.0, -- multiplier to damage for druid abilities
		["subtlety"] = 1.0, -- multiplier to healing threat. e.g. "0.8" for all talents.
		["ferocity"] = 0.0, -- modifier to rage cost of maul and swipe. e.g. "-5" for all talents.
		["tranquilitythreat"] = 1.0 -- multiplier, from Improved Tranquility Talent
	},
	["mage"] = 
	{
		["arcanist"] = false, -- 8 piece bonus
		["netherwind"] = false, -- 3 piece
		["arcanethreat"] = 0.0, -- almost a modifier to global threat (spells only) e.g. "-0.4" for 2/2 talent points.
		["frostthreat"] = 0.0, -- almost a modifier to global threat (spells only) e.g. "-0.3" for 3/3 talent points.
		["firethreat"] = 0.0, -- almost a modifier to global threat (spells only) e.g. "-0.3" for 2/2 talent points.
	},
	["warlock"] = 
	{
		["masterdemo"] = 0.0, -- modifier to global threat. e.g. "-0.2" for imp out, 5/5.
		["nemesis"] = false, -- 8 piece bonus
	},
	["priest"] =
	{
		["silentresolve"] = 0.0, -- almost a modifier to global threat (spells only) e.g. "-0.2" for fully talented.
		["shadowaffinity"] = 1.0, -- multiplier for shadow damage. e.g. "0.8".
	},
	["paladin"] = 
	{
		["righteousfury"] = 1.0, -- multiplier for threat from holy damage / abilities.
		["healing"] = 0.5, -- this is actually constant, but really fits best in this variable.
	},
	["shaman"] = 
	{
		["rockbiter"] = 0, -- current rank of rockbiter that is active, 0 for none.
		["healing"] = 1.0, -- multiplier. e.g. 0.85 for 3/3 Healing Grace.
	}
}

me.redomods = function()
	
	-- talent / gear searching
	local info
	local rank
	local key
	local value
	
	-- warrior
	if UnitClass("player") == mod.string.get("class", "warrior") then

		-- might 8/8
		if mod.data.getsetpieces(mod.string.get("sets", "might")) == 8 then
			me.mods.warrior.sunderthreat = 1.15
		else
			me.mods.warrior.sunderthreat = 1.0
		end

		-- improved sunder armor
		rank = mod.data.gettalentrank(mod.string.get("spell", "sunder"))
		me.mods.warrior.sundercost = -rank

		-- improved heroic strike
		rank = mod.data.gettalentrank(mod.string.get("spell", "heroicstrike"))
		me.mods.warrior.heroicstrikecost = -rank

		-- defiance
		rank = mod.data.gettalentrank(mod.string.get("talent", "defiance"))
		me.mods.warrior.defiance = 0.03 * rank

		-- impale
		rank = mod.data.gettalentrank(mod.string.get("talent", "impale"))
		me.mods.warrior.impale = 0.05 * rank

	-- rogue
	elseif UnitClass("player") == mod.string.get("class", "rogue") then
		
		if mod.data.getsetpieces(mod.string.get("sets", "bloodfang")) > 4 then
			me.mods.rogue.feint = 1.25
		else
			me.mods.rogue.feint = 1.0
		end
	
	-- priest
	elseif UnitClass("player") == mod.string.get("class", "priest") then
		
		-- silent resolve
		rank = mod.data.gettalentrank(mod.string.get("talent", "silentresolve")) 
		me.mods.priest.silentresolve = -0.04 * rank
				
		-- shadow affinity
		rank = mod.data.gettalentrank(mod.string.get("talent", "shadowaffinity"))
		me.mods.priest.shadowaffinity = 1 - rank * 0.25 / 3
	
	-- druid
	elseif UnitClass("player") == mod.string.get("class", "druid") then
		
		-- subtlety
		rank = mod.data.gettalentrank(mod.string.get("talent", "druidsubtlety"))
		me.mods.druid.subtlety = 1 - 0.04 * rank
			
		-- feral instinct
		rank = mod.data.gettalentrank(mod.string.get("talent", "feralinstinct"))
		me.mods.druid.feralinstinct = 0.03 * rank
		
		-- ferocity
		rank = mod.data.gettalentrank(mod.string.get("talent", "ferocity"))
		me.mods.druid.ferocity = -rank
		
		-- savage fury
		rank = mod.data.gettalentrank(mod.string.get("talent", "savagefury"))
		me.mods.druid.savagefury = rank * 0.1
		
		-- improved tranquility
		rank = mod.data.gettalentrank(mod.string.get("talent", "tranquility"))
		me.mods.druid.tranquilitythreat = 1.0 - rank * 0.4
		
	-- mage
	elseif UnitClass("player") == mod.string.get("class", "mage") then
		
		-- arcanist 8 piece
		if mod.data.getsetpieces(mod.string.get("sets", "arcanist")) == 8 then
			me.mods.mage.arcanist = true
		else
			me.mods.mage.arcanist = false
		end
		
		-- arcane subtelty
		rank = mod.data.gettalentrank(mod.string.get("talent", "arcanesubtlety"))
		me.mods.mage.arcanethreat = 0.0 - 0.2 * rank

		-- frost channeling
		rank = mod.data.gettalentrank(mod.string.get("talent", "frostchanneling"))
		me.mods.mage.frostthreat = 0.0 - 0.1 * rank
		
		-- burning soul
		rank = mod.data.gettalentrank(mod.string.get("talent", "burningsoul"))
		me.mods.mage.firethreat = 0.0 - 0.15 * rank
		
		-- netherwind 3 piece
		if mod.data.getsetpieces(mod.string.get("sets", "netherwind")) > 2 then
			me.mods.mage.netherwind = true
		else
			me.mods.mage.netherwind = false
		end
	
	-- warlock
	elseif UnitClass("player") == mod.string.get("class", "warlock") then
		-- master demonologist
		rank = mod.data.gettalentrank(mod.string.get("talent", "masterdemonologist"))
		
		-- check for imp
		if UnitCreatureFamily("pet") == mod.string.get("misc", "imp") then
			me.mods.warlock.masterdemo = -0.04 * rank
		else
			me.mods.warlock.masterdemo = 0
		end
		
		-- nemesis 8 piece
		if mod.data.getsetpieces(mod.string.get("sets", "nemesis")) == 8 then
			me.mods.warlock.nemesis = true
		else
			me.mods.warlock.nemesis = false
		end
		
	-- shaman
	elseif UnitClass("player") == mod.string.get("class", "shaman") then
		
		-- healing grace
		rank = mod.data.gettalentrank(mod.string.get("talent", "healinggrace"))
		me.mods.shaman.healing = 1.0 - 0.05 * rank
		
	-- paladin
	elseif UnitClass("player") == mod.string.get("class", "paladin") then
		
		local multi = 1.0
		
		-- righteous fury buff:
		if mod.data.isbuffpresent("Interface\\Icons\\Spell_Holy_SealOfFury") then
			multi = multi + 0.6
		
			-- talents
			rank = mod.data.gettalentrank(mod.string.get("talent", "righteousfury"))
			multi = multi + (0.5 * rank / 3)
		end
		
		me.mods.paladin.righteousfury = multi
	end
	
end


----------------------------------------------------------------------------------------

----------------------------
--  Global Threat Mods    --
----------------------------

--[[
	There are certain threat modifiers that are applied to all threat done, and that stack additively.
For instance with the "Blessing of Salvation" buff, all your threat is reduced by 30%. Wit the Arcanist
8 piece bonus, threat is reduced by 15%, but these bonuses add directly, so that with both bonuses, your 
threat is reduced by 45%.
	mod.my.globalthreat provides a list of all your global bonuses that are currently active. The data is
designed to be easily printed to the user.
	The data is updated frequently, because the modifiers can come on and off at any time. This means once
every OnUpdate, roughly 40 times per second. As a result we take pains to avoid the creation of new lists,
which would otherwise generate excess heap memory.
	me.globalthreat.modifiers is an ARRAY, i.e. a list keyed by [1], [2], [3], etc. Each item of the array the value
is a list with properties
	["value"] - the fractional reduction in threat. -30% from Blessing of Salvation would be "-0.3".
	["reason"] - a description of the effect, for the user's benefit. e.g. "Blessing of Salvation".
	["isactive"] - if false, this value and all further values should be ignored. Since we don't want to keep
	             recreating lists in a frequently called procedure, when a threat modifier is no longer active,
				 <isactive> is set to false, and that value is ignored. A newly added buff will overwrite the
				other data fields, but set <isactive> to true.
	There is one final part me.globalthreat, a key-value pair. The key is "value" and the value is your total
modifier, i.e. the sum of all the modifiers. For a normal character with no buffs / talents / items, this
value would be 1.0, e.g.
	me.globalthreat = 
{
	["value"] = 1.0,
	["modifiers"] = 
	{
		[1] = 
		{
			["value"] = 1.0,
			["reason"] = "Base Value",
			["isactive"] = true,
		}
	}
}
]]
me.globalthreat = 
{
	["value"] = 0.0,
	["modifiers"] = { }
}

--[[ 
me.modifyglobalthreat(value, reason)
Adds a global threat modifier to me.globalthreat.
<value> and <reason> are as defined in me.globalthreat (see comments above).
This method will first look for a value in the me.globalthreat array that is labeled inactive (unused memory) 
and write the values there. Otherwise it will create a new value and append it to the array.
Heap memory creation occurs a maximum of n times, where n is the largest number of +- threat buffs you get.
This is on the order of 5, and for the whole running time, so negligable.
]]
me.modifyglobalthreat = function(value, reason)
	
	-- update the total / sum
	me.globalthreat.value = me.globalthreat.value + value
	
	-- look for an unused array position
	local x
	local count = table.getn(me.globalthreat.modifiers)
	
	for x = 1, count do
		
		if me.globalthreat.modifiers[x].isactive == false then
			
			-- found an inactive array index. Activate it and write values
			me.globalthreat.modifiers[x].value = value
			me.globalthreat.modifiers[x].reason = reason
			me.globalthreat.modifiers[x].isactive = true 
			return
		end
	end
	
	-- all the slots in the array are being used. Add a new value to the end
	table.insert(me.globalthreat.modifiers, {["value"] = value, ["reason"] = reason, ["isactive"] = true})
	
end

--[[ 
me.redoglobalthreat()
Recalculates all your global threat modifiers.
]]
me.redoglobalthreat = function()

	-- 1) reset
	local x
	for x = 1, table.getn(me.globalthreat.modifiers) do
		me.globalthreat.modifiers[x].isactive = false
	end
	
	me.globalthreat.value = 0.0

	-- 2) rebuild
	me.modifyglobalthreat(1.0, mod.string.get("threatmod", "basevalue"))
	
	-- Tranquil Air
	if mod.data.isbuffpresent("Interface\\Icons\\Spell_Nature_Brilliance") == true then
		me.modifyglobalthreat(-0.2, mod.string.get("threatmod", "tranquilair"))
	end
	
	-- Blessing of Salvation
	if mod.data.isbuffpresent("Interface\\Icons\\Spell_Holy_SealOfSalvation") == true then
		me.modifyglobalthreat(-0.3, mod.string.get("threatmod", "salvation"))
		
	elseif mod.data.isbuffpresent("Interface\\Icons\\Spell_Holy_GreaterBlessingofSalvation") == true then
		me.modifyglobalthreat(-0.3, mod.string.get("threatmod", "salvation"))
	end
	
	-- Burning Adrenaline
	if me.states["burningadrenaline"].value == true then
		me.modifyglobalthreat(-0.75, mod.string.get("spell", "burningadrenaline"))
	end
	
	-- Fetish of the Sand Reaver
	if me.states["arcaneshroud"].value == true then
		me.modifyglobalthreat(-0.7, mod.string.get("spell", "arcaneshroud"))
	end
	
	local linkstring
	local enchant
	
	-- +2% threat enchant to gloves
	linkstring = GetInventoryItemLink("player", 10) or "" 
	enchant = string.gsub(linkstring, ".-item:%d+:(%d+).*", "%1", 1)
	
	if enchant == "2613" then
		me.modifyglobalthreat(0.02, mod.string.get("threatmod", "glovethreatenchant"))
	end
	
	-- -2% threat enchant to back
	linkstring = GetInventoryItemLink("player", 15) or ""
	enchant = string.gsub(linkstring, ".-item:%d+:(%d+).*", "%1", 1)
	
	if enchant == "2621" then
		me.modifyglobalthreat(-0.02, mod.string.get("threatmod", "backthreatenchant"))
	end
	
	local stance
	
	-- Warrior
	if UnitClass("player") == mod.string.get("class", "warrior") then
		stance = me.getstanceindex()
		
		if stance == 1 then
			me.modifyglobalthreat(-0.2, mod.string.get("threatmod", "battlestance"))
			
		elseif stance == 2 then
			me.modifyglobalthreat(0.3, mod.string.get("threatmod", "defensivestance"))
			me.modifyglobalthreat(me.mods.warrior.defiance, mod.string.get("talent", "defiance"))
	
		elseif stance == 3 then
			me.modifyglobalthreat(-0.2, mod.string.get("threatmod", "berserkerstance"))
		end
		
	-- Druid Stances
	elseif UnitClass("player") == mod.string.get("class", "druid") then
		stance = me.getstanceindex()
		
		if stance == 1 then
			me.modifyglobalthreat(0.3, mod.string.get("threatmod", "bearform"))
			me.modifyglobalthreat(me.mods.druid.feralinstinct, mod.string.get("talent", "feralinstinct"))
		end
	
	-- Rogue
	elseif UnitClass("player") == mod.string.get("class", "rogue") then
		me.modifyglobalthreat(-0.2, mod.string.get("class", "rogue"))
		
	-- Arcanist
	elseif (UnitClass("player") == mod.string.get("class", "mage")) and (me.mods.mage.arcanist == true) then
		me.modifyglobalthreat(-0.15, mod.string.get("sets", "arcanist") .. " 8/8")
		
	-- Warlock
	elseif (UnitClass("player") == mod.string.get("class", "warlock")) and (me.mods.warlock.masterdemo ~= 0) then
		me.modifyglobalthreat(me.mods.warlock.masterdemo, mod.string.get("talent", "masterdemonologist"))
	end	
end


--[[ 
me.getstanceindex()
For Druids and Warriors, returns an integer saying which stance you are in.
]]
me.getstanceindex = function()
	
	local x = 0
	local isactive
	local texture 
	
	while true do
		x = x + 1
		texture, _, isactive = GetShapeshiftFormInfo(x)
		
		if texture == nil then
			return 0
		end
		
		if isactive == 1 then
			return x
		end
	end
	
end